import numpy as np
from prefect import task
from sklearn.decomposition import PCA


@task
def fit_pca_model(features: np.ndarray, components: int, ordering: np.ndarray) -> PCA:
    # Fit data.
    pca = PCA(n_components=components)
    pca = pca.fit(features)

    # Reorient features by ordering data.
    transformed = pca.transform(features)
    for i in range(components):
        pearson = np.corrcoef(ordering, transformed[:, i])
        if pearson[0, 1] < 0:
            pca.components_[i] = pca.components_[i] * -1

    return pca



import xml.etree.ElementTree as ET

from prefect import task


@task
def compile_shape_modes(
    shape_modes: dict,
    views: list[str],
    regions: list[str],
    variance: list[float],
    colors: dict,
    box: tuple[int, int],
    scale: float,
) -> str:
    # Extract layout parameters
    width, height = box
    components = sorted(list(shape_modes.keys()))
    points = sorted(list(shape_modes[components[0]].keys()))

    # Initialize output
    root = ET.fromstring("<svg></svg>")
    slice_group = add_group(root, "slices", f"translate({width},{height / 2})")
    row_group = add_group(root, "row_labels", f"translate(0,{height / 2})")
    col_group = add_group(root, "col_labels", f"translate({width},0)")

    # Add row labels
    for component_index, component in enumerate(components):
        transform = f"translate(0,{component_index * height})"
        add_row_labels(row_group, transform, component, variance[component_index], width, height)

    # Add col labels
    for point_index, point in enumerate(points):
        transform = f"translate({point_index * width * len(views)},0)"
        add_col_labels(col_group, transform, f"{point}σ", height, width, views)

    # Insert shape modes
    for component_index, component in enumerate(components):
        component_id = f"component[{component}]"
        component_group = add_group(
            slice_group, component_id, f"translate(0,{component_index * height})"
        )

        for point_index, point in enumerate(points):
            point_id = f"point[{point}]"
            point_group = add_group(
                component_group,
                f"{component_id}_{point_id}",
                f"translate({point_index * width * len(views)},0)",
            )

            for view_index, view in enumerate(views):
                view_transform = f"translate({view_index * width},0)"
                view_group = add_group(
                    point_group, f"{component_id}_{point_id}_{view}", view_transform
                )
                add_border(view_group, width, height)

                for region in regions:
                    rotate = 0 if view == "top" else 90
                    color = colors[region]
                    append_svg_element(
                        shape_modes[component][point][region][view],
                        view_group,
                        width,
                        height,
                        scale,
                        rotate,
                        color,
                    )

    # Clean namespaces
    clear_svg_namespaces(root)
    for element in root.findall(".//*"):
        clear_svg_namespaces(element)

    # Set SVG size and namespace
    root.set("xmlns", "http://www.w3.org/2000/svg")
    root.set("height", str(len(components) * height + height / 2))
    root.set("width", str(len(points) * width * len(views) + width))

    return ET.tostring(root, encoding="unicode")


def add_group(root: ET.Element, group_id: str, group_transform: str) -> ET.Element:
    return ET.SubElement(root, "g", {"id": group_id, "transform": group_transform})


def add_border(root: ET.Element, width: int, height: int) -> None:
    ET.SubElement(
        root,
        "rect",
        {
            "width": str(width),
            "height": str(height),
            "x": "0",
            "y": "0",
            "fill": "none",
            "stroke": "#eeeeee",
            "stroke-width": "0.5",
        },
    )


def add_col_labels(
    root: ET.Element, transform: str, text: str, height: int, width: int, views: list[str]
) -> None:
    group = ET.SubElement(root, "g", {"transform": transform})

    ET.SubElement(
        group,
        "text",
        {
            "font-family": "Helvetica",
            "font-size": "16pt",
            "text-anchor": "middle",
            "font-weight": "bold",
            "x": str(len(views) * width / 2),
            "y": str(20),
        },
    ).text = text

    for i, view in enumerate(views):
        ET.SubElement(
            group,
            "text",
            {
                "font-family": "Helvetica",
                "font-size": "14pt",
                "text-anchor": "middle",
                "font-style": "italic",
                "x": str((i + 0.5) * width),
                "y": str(height / 2 - 10),
            },
        ).text = f"{view.replace('_', ' ')} view"


def add_row_labels(
    root: ET.Element, transform: str, component: int, variance: float, width: int, height: int
) -> None:
    group = ET.SubElement(root, "g", {"transform": transform})

    ET.SubElement(
        group,
        "text",
        {
            "font-family": "Helvetica",
            "font-size": "16pt",
            "text-anchor": "middle",
            "font-weight": "bold",
            "x": str(width / 2),
            "y": str(height / 2 - 12),
        },
    ).text = "Shape"

    ET.SubElement(
        group,
        "text",
        {
            "font-family": "Helvetica",
            "font-size": "16pt",
            "text-anchor": "middle",
            "font-weight": "bold",
            "x": str(width / 2),
            "y": str(height / 2 + 8),
        },
    ).text = f"Mode {component}"

    ET.SubElement(
        group,
        "text",
        {
            "font-family": "Helvetica",
            "font-size": "14pt",
            "text-anchor": "middle",
            "x": str(width / 2),
            "y": str(height / 2 + 30),
        },
    ).text = f"{variance*100:.1f}%"


def append_svg_element(
    element: str,
    root: ET.Element,
    width: int,
    height: int,
    scale: float = 1.0,
    rotate: int = 0,
    color: str = "#555",
) -> None:
    """Append svg element to root."""
    path = ET.fromstring(element).findall(".//")[0]

    cx = width / 2
    cy = height / 2

    path.set("fill", "none")
    path.set("stroke", color)
    path.set("stroke-width", str(width / 80 / scale))
    path.set("transform", f"rotate({rotate},{cx},{cy}) translate({cx},{cy}) scale({scale})")
    root.insert(0, path)


def clear_svg_namespaces(svg: ET.Element) -> None:
    _, has_namespace, postfix = svg.tag.partition("}")
    if has_namespace:
        svg.tag = postfix



import tempfile

import numpy as np
import pandas as pd
import trimesh
from aicsshparam import shtools
from prefect import task
from sklearn.decomposition import PCA
from vtkmodules.vtkCommonDataModel import vtkPolyData
from vtkmodules.vtkIOPLY import vtkPLYWriter


@task
def extract_shape_modes(
    pca: PCA, data: pd.DataFrame, components: int, regions: list[str], order: int, delta: float
) -> dict:
    features = data.filter(like="shcoeffs").columns.values
    data_transform = pca.transform(data[features].values)
    means = data_transform.mean(axis=0)
    stds = data_transform.std(axis=0, ddof=1)

    shape_svgs = {}

    for component in range(components):
        point_vector = np.zeros((components))
        component_shape_modes = {}

        for point in np.arange(-2, 2.5, delta):
            point_vector[component] = point
            vector = means + np.multiply(stds, point_vector)
            point_shape_modes = {}

            for region in regions:
                shape_mode_slices = extract_shape_mode_slices(pca, vector, features, order, region)
                point_shape_modes[region] = shape_mode_slices

            component_shape_modes[point] = point_shape_modes

        shape_svgs[component + 1] = component_shape_modes

    return shape_svgs


def extract_shape_mode_slices(
    pca: PCA, vector: np.ndarray, feature_names: list[str], order: int, region: str
) -> dict:
    prefix = ""
    suffix = f".{region}" if region != "DEFAULT" else ""
    mesh = construct_mesh_from_points(pca, vector, feature_names, order, prefix, suffix)
    mesh = convert_vtk_to_trimesh(mesh)
    slices = get_mesh_slices(mesh)
    return slices


def construct_mesh_from_points(
    pca: PCA,
    points: np.ndarray,
    feature_names: list[str],
    order: int,
    prefix: str = "",
    suffix: str = "",
) -> vtkPolyData:
    """Constructs mesh given PCA transformation points."""
    coeffs = pd.Series(pca.inverse_transform(points), index=feature_names)
    coeffs_map = np.zeros((2, order + 1, order + 1), dtype=np.float32)

    for l in range(order + 1):
        for m in range(order + 1):
            coeffs_map[0, l, m] = coeffs[f"{prefix}shcoeffs_L{l}M{m}C{suffix}"]
            coeffs_map[1, l, m] = coeffs[f"{prefix}shcoeffs_L{l}M{m}S{suffix}"]

    mesh, _ = shtools.get_reconstruction_from_coeffs(coeffs_map)
    return mesh


def convert_vtk_to_trimesh(mesh: vtkPolyData) -> trimesh.Trimesh:
    with tempfile.NamedTemporaryFile() as temp:
        writer = vtkPLYWriter()
        writer.SetInputData(mesh)
        writer.SetFileTypeToASCII()
        writer.SetFileName(f"{temp.name}.ply")
        _ = writer.Write()
        mesh = trimesh.load(f"{temp.name}.ply")

    return mesh


def get_mesh_slices(mesh: trimesh.Trimesh) -> dict:
    return {
        "side_1": get_mesh_slice(mesh, (0, 1, 0)),
        "side_2": get_mesh_slice(mesh, (1, 0, 0)),
        "top": get_mesh_slice(mesh, (0, 0, 1)),
    }


def get_mesh_slice(mesh: trimesh.Trimesh, normal: tuple[int, int, int]) -> str:
    """Get svg slice of mesh along plane for given normal."""
    mesh_slice = mesh.section_multiplane(mesh.centroid, normal, [0])
    svg = trimesh.path.exchange.svg_io.export_svg(mesh_slice[0])
    return svg



import numpy as np
from prefect import task


@task
def make_voxels_array(voxels: list[list], scale: int = 1) -> np.ndarray:
    # Center voxels around (0,0,0).
    center_x, center_y, center_z = [round(x) for x in np.array(voxels).mean(axis=0)]
    all_xyz_centered = [(z - center_z, y - center_y, x - center_x) for x, y, z in voxels]

    # Create empty array.
    mins = np.min(all_xyz_centered, axis=0)
    maxs = np.max(all_xyz_centered, axis=0)
    height, width, length = np.subtract(maxs, mins) + 3
    array = np.zeros((height, width, length), dtype=np.uint8)

    # Fill in voxel array.
    all_xyz_offset = [
        (z - mins[0] + 1, y - mins[1] + 1, x - mins[2] + 1) for z, y, x in all_xyz_centered
    ]
    vals = [1] * len(all_xyz_offset)
    array[tuple(np.transpose(all_xyz_offset))] = vals

    # Scale the array.
    if scale > 1:
        array = array.repeat(scale, axis=0).repeat(scale, axis=1).repeat(scale, axis=2)

    return array



import xml.etree.ElementTree as ET

from prefect import task

from abm_shape_collection.compile_shape_modes import (
    add_border,
    add_col_labels,
    add_group,
    add_row_labels,
    append_svg_element,
    clear_svg_namespaces,
)


@task
def merge_shape_modes(
    shape_modes: dict,
    views: list[str],
    regions: list[str],
    variance: list[float],
    colors: dict,
    box: tuple[int, int],
    scale: float,
) -> str:
    # Extract layout parameters
    width, height = box
    components = sorted(list(shape_modes.keys()))
    points = sorted(list(shape_modes[components[0]].keys()), reverse=True)

    # Initialize output
    root = ET.fromstring("<svg></svg>")
    slice_group = add_group(root, "slices", f"translate({width},{height / 2})")
    row_group = add_group(root, "row_labels", f"translate(0,{height / 2})")
    col_group = add_group(root, "col_labels", f"translate({width},0)")

    # Add row labels
    for component_index, component in enumerate(components):
        transform = f"translate(0,{component_index * height})"
        add_row_labels(row_group, transform, component, variance[component_index], width, height)

    # Add col labels
    for region_index, region in enumerate(regions):
        transform = f"translate({region_index * width * len(views)},0)"
        add_col_labels(col_group, transform, region, height, width, views)

    # Insert shape modes
    for component_index, component in enumerate(components):
        component_id = f"component[{component}]"
        component_group = add_group(
            slice_group, component_id, f"translate(0,{component_index * height})"
        )

        for region_index, region in enumerate(regions):
            region_id = f"region[{region}]"
            region_group = add_group(
                component_group,
                f"{component_id}_{region_id}",
                f"translate({region_index * width * len(views)},0)",
            )

            for view_index, view in enumerate(views):
                view_transform = f"translate({view_index * width},0)"
                view_group = add_group(
                    region_group, f"{component_id}_{region_id}_{view}", view_transform
                )
                add_border(view_group, width, height)

                for point in points:
                    rotate = 0 if view == "top" else 90
                    color = "#ddd" if point == 0 else shade_color(colors[region], -point / 4)

                    append_svg_element(
                        shape_modes[component][point][region][view],
                        view_group,
                        width,
                        height,
                        scale,
                        rotate,
                        color,
                    )

    # Clean namespaces
    clear_svg_namespaces(root)
    for element in root.findall(".//*"):
        clear_svg_namespaces(element)

    # Set SVG size and namespace
    root.set("xmlns", "http://www.w3.org/2000/svg")
    root.set("height", str(len(components) * height + height / 2))
    root.set("width", str(len(regions) * width * len(views) + width))

    return ET.tostring(root, encoding="unicode")


def shade_color(color: str, alpha: float) -> str:
    old_color = color.replace("#", "")
    old_red, old_green, old_blue = [int(old_color[i : i + 2], 16) for i in (0, 2, 4)]
    layer_color = 0 if alpha < 0 else 255

    new_red = round(old_red + (layer_color - old_red) * abs(alpha))
    new_green = round(old_green + (layer_color - old_green) * abs(alpha))
    new_blue = round(old_blue + (layer_color - old_blue) * abs(alpha))

    return f"#{new_red:02x}{new_green:02x}{new_blue:02x}"



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['abm_shape_collection']

package_data = \
{'': ['*']}

install_requires = \
['aicsshparam>=0.1.1,<0.2.0',
 'numpy>=1.23.4,<2.0.0',
 'pandas>=1.5.1,<2.0.0',
 'prefect>=2.6.0,<3.0.0',
 'trimesh>=3.18.3,<4.0.0',
 'vtk>=9.2.5,<10.0.0']

setup_kwargs = {
    'name': 'abm-shape-collection',
    'version': '0.2.0',
    'description': 'Collection of tasks for analyzing cell shapes.',
    'long_description': '[![Build Status](https://github.com/allen-cell-animated/abm-shape-collection/workflows/build/badge.svg)](https://github.com/allen-cell-animated/abm-shape-collection/actions?query=workflow%3Abuild)\n[![Codecov](https://img.shields.io/codecov/c/gh/allen-cell-animated/abm-shape-collection?token=FMO7XIX5KH)](https://codecov.io/gh/allen-cell-animated/abm-shape-collection)\n[![Lint Status](https://github.com/allen-cell-animated/abm-shape-collection/workflows/lint/badge.svg)](https://github.com/allen-cell-animated/abm-shape-collection/actions?query=workflow%3Alint)\n[![Documentation](https://github.com/allen-cell-animated/abm-shape-collection/workflows/documentation/badge.svg)](https://allen-cell-animated.github.io/abm-shape-collection/)\n[![Code style](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n',
    'author': 'Jessica S. Yu',
    'author_email': 'jesyu@uw.edu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from .calculate_shape_stats import calculate_shape_stats
from .calculate_size_stats import calculate_size_stats
from .compile_shape_modes import compile_shape_modes
from .extract_shape_modes import extract_shape_modes
from .fit_pca_model import fit_pca_model
from .get_shape_coefficients import get_shape_coefficients
from .make_voxels_array import make_voxels_array
from .merge_shape_modes import merge_shape_modes
